#ifndef _Certificate_order_h
#define _Certificate_order_h 1

#include <sys/time.h>
#include "types.h"
#include "Time.h"
#include "parameters.h"
#include "Bitmap.h"

template <class T> class Certificate_order {
  //
  // A certificate is a set of "matching" messages from different
  // replicas. 
  //
  // T must have the following methods:
  // bool match(T*);
  // // Effects: Returns true iff the messages match
  //
  // int id();
  // // Effects: Returns the identifier of the principal that
  // // sent the message.
  //
  // bool verify();
  // // Effects: Returns true iff the message is properly authenticated
  // // and statically correct.
  //
  // bool full();
  // // Effects: Returns true iff the message is full
  //
  // bool encode(FILE* o);
  // bool decode(FILE* i);
  // Effects: Encodes and decodes object state from stream. Return
  // true if successful and false otherwise.
  
public:
  Certificate_order(int complete=0, int correct=0, int maxs=0);
  // Requires: "complete" (2*f()+1) and "correct" (f()+1) to be specified 
  // Effects: Creates an empty certificate. The certificate is
  // complete when it contains at least "complete" matching messages
  // from different replicas. If the complete argument is omitted (or
  // 0) it is taken to be 2f+1. The certificate is correct if it received 
  // matching "correct" messages. If the complete argument is omitted (or
  // 0) it is taken to be 2f+1. If the correct is ommitted it is taken 
  // to be f()+1. "maxs" is the total number of messages that can be 
  // received and it is set to 3*f()+1 if not specified.
  // 

  ~Certificate_order();
  // Effects: Deletes certificate and all the messages it contains.
 
  bool add(T *m);
  // Effects: Adds "m" to the certificate and returns true provided
  // "m" satisfies:
  // 1. there is no message from "m.id()" in the certificate
  // 2. "m->verify() == true"
  // 3. if "cvalue() != 0", "cvalue()->match(m)";
  // otherwise, it has no effect on this and returns false.  This
  // becomes the owner of "m" (i.e., no other code should delete "m"
  // or retain pointers to "m".)

  T *cvalue() const;
  // Effects: Returns the correct message value for this certificate
  // or 0 if this value is not known. Note that the certificate
  // retains ownership over the returned value (e.g., if clear or
  // mark_stale are called the value may be deleted.)

  T *cvalue_clear();
  // Effects: Returns the correct message value for this certificate
  // or 0 if this value is not known. If it returns the correct value,
  // it removes the message from the certificate and clears the
  // certificate (that is the caller gains ownership over the returned
  // value.)

  int num_correct() const;
  // Effects: Returns the number of messages with the correct value
  // in this.

  bool is_complete() const;
  // Is certificate complete 

  int num_locally_committed();
  // Checks to see if atleast num_complete() replies have last_committed > seq_num
  // This is used for read-only optimization

  void update_locally_committed();
  // Update locally committed whenever new message is inserted

  bool is_locally_committed();
  // if num_complete() number of replicas all committed this request

  int num_complete() const;
  // Is certificate complete 

  int num_elements();
  // Number of elements added

  bool is_done() const;
  // Is certificate complete 

  bool is_recv_complete() const;
  // Have we received messages from all replicas

  bool max_match() const;
  // Do we have a certificate with max match 
  
  int msize() const;
  // Return max entries that can be in a certificate 

  void make_complete();
  // Effects: If cvalue() is not null, makes the certificate
  // complete.
    
  void mark_stale();
  // Effects: Discards all messages in certificate except mine. 

  void clear(); 
  // Effects: Discards all messages in certificate

  bool is_empty() const;
    
  class Message_val_order {
  public:
    T *m;
    int count; 
    int lc_count; // locally_committed count
    int lc_last_checked; // Index that is last checked in this group for local commit

    int *id_list; 
    T **reply_list;
    inline Message_val_order() { m = 0; count = 0; id_list = NULL; reply_list = NULL;}
    inline void clear() { 
      m = 0;
      for (int j=0; j<count;j++) {
         reply_list[j] = NULL;
         id_list[j] = -1;
      }
      count = 0;
      lc_count = 0;
      lc_last_checked = 0;
    }

    inline ~Message_val_order() { 
      m = 0;
      if (reply_list != NULL) free(reply_list);
      if (id_list != NULL) free(id_list);
      count = 0;
      lc_count = 0;
      lc_last_checked = 0;
   }
    
  };

  bool collect_proof(struct iovec *&p, int& len, int& size);
  // Collect proof 

private:
 
  Message_val_order *vals;     // Vector with all distinct message values in this
  Message_val_order *all_vals; // Vector with all (valid) messages received that is used  for generating proof when
                         // the OR certificate is incomplete

  int *all_id_list;      // id list of all the valid messages received 
  Bitmap bmap;           // bitmap with replicas whose message is in this.
  int next_elem;         // Number of elements in the bmap and all_vals

  
  int max_size;          // maximum number of elements in vals, f+1
  int cur_size;          // current number of elements in vals
  int total_recv;        // total number of valid OR message 

  int correct;           // value is correct if it appears in at least "correct" messages
  Message_val_order *c;        // correct certificate value or 0 if unknown.

  int complete;          // certificate is complete if "num_correct() >= complete"

  T *mym;                // my message in this or null if I have no message in this 
  Time t_sent;           // time at which mym was last sent

  // The implementation assumes:
  // correct > 0 and complete > correct 
};

template <class T> 
inline  int Certificate_order<T>::num_elements() { return next_elem;}

template <class T> 
inline T *Certificate_order<T>::cvalue() const { return (c) ? c->m : 0; }

template <class T> 
inline int Certificate_order<T>::num_correct() const {  return (c) ? c->count : 0; }

template <class T> 
inline int Certificate_order<T>::num_locally_committed()  {  return (c) ? c->lc_count : 0;}

template <class T> 
inline bool Certificate_order<T>::is_locally_committed() {  return c->lc_count >= complete;}

template <class T>
inline bool Certificate_order<T>::is_complete() const { return num_correct() >= complete; }

template <class T>
inline int Certificate_order<T>::num_complete() const { return complete; }

template <class T>
inline bool Certificate_order<T>::is_recv_complete() const { return next_elem >= complete; }

template <class T>
inline bool Certificate_order<T>::max_match() const { return (c) ? (c->count == max_size) : false; }

template <class T>
inline int Certificate_order<T>::msize() const { return max_size; }

// is done is used when all the messages are received from different replicas
template <class T>
inline bool Certificate_order<T>::is_done() const { return (c) ? (c->count+cur_size == max_size) : (cur_size == max_size);}


template <class T>
inline void Certificate_order<T>::make_complete() {
  if (c) {
    c->count = complete;
  }
}

template <class T>
inline void Certificate_order<T>::clear() {
  for (int i=0; i < cur_size; i++) {
    //fprintf(stderr,"Clearing %x with id :%d seq no: %d in vals  with count :%d total :%d \n",
    //vals[i].m, vals[i].m->id(),vals[i].m->seqno(),vals[i].count, cur_size); 
    vals[i].clear();
  }

  for (int i=0; i < next_elem; i++) {
    // delete all stored OR-requests
    //fprintf(xstderr,"deleting %x with id :%d seq no: %d in all vals \n",
    //all_vals[i].m, all_vals[i].m->id(),all_vals[i].m->seqno()); 
    delete all_vals[i].m;
  }

  bmap.clear();
  c = 0;
  mym = 0;
  t_sent = 0;
  cur_size = 0;
  next_elem = 0;
}

#endif // Certificate_order_h
  










