#ifndef _Committed_locally_h
#define _Committed_locally_h 1

#include "types.h"
#include "Message.h"
#include "Digest.h"

class Principal;
class Rep_info;

// 
// Committed_locally messages have the following format.
//
struct Committed_locally_rep : public Message_rep {
  View v;                // current view
  Request_id rid;        // unique request identifier
  Digest rh_digest;      // request history digest
  int replica;           // id of replica sending the reply

  // Followed by a MAC authenticating that request is committed locally
};

class Committed_locally : public Message {
  // 
  // Committed_locally messages
  //
public:
  Committed_locally() : Message() {}

  Committed_locally(Committed_locally_rep *r);

  Committed_locally(View view, Request_id req, int replica, Digest &rh_d, Principal *clientp);
  // Return reply to "clientp" with history digest "rh_d", reply digest, "rep_d"
  // request rid "req"

  void authenticate(Principal *p);
  // Effects: Terminates the construction of a reply message by
  // setting the length of the reply to "act_len", appending a MAC,
  // and trimming any surplus storage.
  
  View view() const;
  // Effects: Fetches the view from the message

  Request_id request_id() const;
  // Effects: Fetches the request identifier from the message.

  int id() const;
  // Effects: Fetches the reply identifier from the message.

  Digest &request_history_digest() const;
  // Effects : Fetches the request-history digest from the message

  bool verify();
  // Effects: Verifies if the message is authenticated by rep().replica.

  bool match(Committed_locally *r);
  // Effects: Returns true if the replies match.

  bool match_rh_digest(Digest rh_digest);
  // Effects : Returns true if the request history digest in the reply 
  // matches with rh_digest

   static bool convert(Message *m1, Committed_locally *&m2);
  // Effects: If "m1" has the right size and tag of a "Committed_locally", casts
  // "m1" to a "Committed_locally" pointer, returns the pointer in "m2" and
  // returns true. Otherwise, it returns false. Convert also trims any
  // surplus storage from "m1" when the conversion is successfull.

private:
  Committed_locally_rep &rep() const;
  // Effects: Casts "msg" to a Committed_locally_rep&
  
};

inline Committed_locally_rep& Committed_locally::rep() const { 
  th_assert(ALIGNED(msg), "Improperly aligned pointer");
  return *((Committed_locally_rep*)msg); 
}
   
inline View Committed_locally::view() const { return rep().v; }

inline Request_id Committed_locally::request_id() const { return rep().rid; }

inline int Committed_locally::id() const { return rep().replica; }

inline Digest& Committed_locally::request_history_digest() const { return rep().rh_digest; }

  
inline bool Committed_locally::match(Committed_locally *r) {
  return ((rep().rh_digest == r->rep().rh_digest) &&  (view() == r->view()) && (rep().rid == r->rep().rid));
}

inline bool Committed_locally::match_rh_digest(Digest rh_digest) {
  return (rep().rh_digest == rh_digest);
}


#endif // _Committed_locally_h
