#include <strings.h>
#include "th_assert.h"
#include "Message_tags.h"
#include "Local_commit.h"
#include "Node.h"
#include "Client.h"
#include "Principal.h"
#include "Order_request.h"
#include "Message.h"
#include "Statistics.h"


Local_commit::Local_commit(bool complete, int tsize, int nelem, Order_request_rep *m) : 
  Message(Local_commit_tag, Max_message_size) {
    if (complete == true) {
	rep().v = m->v; 
        rep().rid = ((Client*)node)->get_rid(); 
        rep().cid = node->id();
        rep().seq_num = m->seq_num;
        rep().rh_digest = m->rh_digest;
        rep().tsize = tsize;
	rep().nelem = nelem;
        rep().extra = 1;
    }
    else {
	rep().v = node->view(); 
        rep().rid = ((Client*)node)->get_rid();
        rep().cid = node->id();
        rep().seq_num = -1;
        rep().tsize = tsize;
	rep().nelem = nelem;
        rep().extra = 0;
    }
}


Local_commit::Local_commit(Local_commit_rep *r) : Message(r) {}


void Local_commit::authenticate(bool use_auth) {
  th_assert((unsigned) node->auth_size() < msize()-sizeof(Local_commit_rep),
	    "Insufficient memory");
  int old_size = sizeof(Local_commit_rep);

  if (use_auth) {
     // Use authenticator 
    node->gen_auth_in(contents(), old_size, contents()+old_size);
  }
  else {
    // Use signature
    rep().extra |= 2;
    node->gen_signature(contents(), old_size, contents()+old_size);
  } 
  // trim(); Don't trim here as we append proof later
}

bool Local_commit::verify() {
const int nid = node->id();
  const int client_id = cid();
  const int old_size = sizeof(Local_commit_rep);
  Principal* p = node->i_to_p(client_id);


  if (p != 0) {
    // 1.0 First check is properly authenticated      
    if (client_id != nid && client_id >= node->n() && size()-old_size >= node->auth_size()) {
      if (!node->verify_auth_out(client_id, contents(), sizeof(Local_commit_rep),
				 contents()+old_size)) {
	return false;
      }
    } else {
      // Message is signed.
      if (size() - old_size >= p->sig_size()) {
	if (p->verify_signature(contents(), sizeof(Local_commit_rep),
				contents()+old_size, true)) {
	  return false;
	}
      }
    }
  }
  return true;
}

bool Local_commit::append_proof(struct iovec *p, int len, int psize) {
  int cid = node->id();
  int old_size = sizeof(Local_commit_rep)+ node->auth_size(cid);
  th_assert((unsigned)psize < (unsigned)(msize()-old_size),"Insufficient memory");

  for (int i=0; i< len; i++) {
    memcpy(contents()+old_size,p[i].iov_base, p[i].iov_len);
    old_size+=p[i].iov_len; 
  }

  set_size(old_size);
  return true;
}

    
 
bool Local_commit::verify(int &match_count) {

  const int nid = node->id();
  const int client_id = cid();
  const int old_size = sizeof(Local_commit_rep);
  Principal* p = node->i_to_p(client_id);
  match_count = 0;
  
  if (p != 0) {
    // 1.0 First check is properly authenticated      
    if (client_id != nid && client_id >= node->n() && size()-old_size >= node->auth_size()) {
      if (!node->verify_auth_out(client_id, contents(), sizeof(Local_commit_rep),
				 contents()+old_size)) {
	return false;
      }
    } else {
      // Message is signed.
      if (size() - old_size >= p->sig_size()) {
	if (p->verify_signature(contents(), sizeof(Local_commit_rep),
				contents()+old_size, true)) {
	  return false;
	}
      }
    }
  }
        
  if ((rep().extra & 1)) {
    
    // Client received valid 2f+1 matching speculative reply messages
    // Check if they are properly authenticated
    int ts = tsize();
    int entries = nelem();
    int reps = entries-2;
   
    if ((ts > 0) && (reps >0)) {
	
      // Check for each of the replicas is properly authenticated
      // Extract the reply header from both the exec replicas and 
      // agreement replicas

      Order_request_rep *r = read_order();
      int *id_list = read_id_list(reps);
      
      //fprintf(stderr,"Included matching responses from %d replicas \n", reps);

      for (int i = 0; i<reps; i++) {
	int rid = id_list[i];
	bool res = false;

	// Check only authenticators from other replicas
	if (rid != nid) {
	  // verify the order request message from rid
	  
	  // Change the replica id of the order request before veryfing authenticator for each replica
	  //fprintf(stderr,"Verifying replica %d \n", rid);
	  r->rep_id = rid;
	  
	  //r.change_replica_id(rid);

	  //res = node->verify_auth_in(rid, (char *)r.contents(), sizeof(Order_request_rep), (char *)(r.contents())+sizeof(Order_request_rep)+i*node->auth_size());
	  res = node->verify_auth_in(rid, (char *)r, sizeof(Order_request_rep), (char *)r+sizeof(Order_request_rep)+i*node->auth_size());
	  if (res) {
	    match_count++;
	    //fprintf(stderr,"Verification passed replica %d \n", rid);
	  }
	}
      }
      //fprintf(stderr,"Verified successfully from %d replicas \n", match_count);
      
    }
  }
  else {
    // TO DO : Client did not receive matching speculative reply message 
    // Next check if there are 2f+1 valid spec. reply messages that do not match
    return false;
  }
  
  return true;
}


bool Local_commit::convert(Message *m1, Local_commit *&m2) {
  if (!m1->has_tag(Local_commit_tag, sizeof(Local_commit_rep)))
    return false;
  
  m1->trim();
  m2 = (Local_commit*)m1;
  return true;
}
