#ifndef _Local_commit_h
#define _Local_commit_h 1

#include "types.h"
#include "Message.h"
#include "Digest.h"
#include "Reply.h"
#include "Order_request.h"
#include "Node.h"

class Principal;
class Rep_info;
class Reply;


// 
// Local_commit messages have the following format.
//
struct Local_commit_rep : public Message_rep {
  View v;                // current view
  int cid;               // Client id 
  Request_id rid;        // unique request identifier
  int seq_num;           // Sequence number  
  Digest rh_digest;      // request history digest
  int tsize;             // Total size of the proof
  int nelem;             // Number of replicas included in the proof

  // 1. Followed by the authenticator signed by the client 

  // 2. Followed by the proof of OR request messages sent by the replicas to the client
};

class Local_commit : public Message {
  // 
  // Local_commit messages
  //
public:
  Local_commit(): Message() {}

  Local_commit(Local_commit_rep *r);

  Local_commit(bool complete, int tsize, int nelem, Order_request_rep *m);
  // Local commit message created by the client. If complete is true 
  // then it includes the proof of 2f+1 matching order-request message digests 
  // in the message with size "psize"

  void authenticate(bool use_mac);
  // Effects: Terminates the construction of a local-commit message by
  // setting the length of the reply to "act_len", appending authenticator/signature,
  // and trimming any surplus storage.
  // if "use_mac" is true then it uses authenticator 
  // otherwise digital signature
  
  
  View view() const;
  // Effects: Fetches the view from the message

  Seqno seqno() const;
  // Returns the seq no of the request 

  Request_id request_id() const;
  // Effects: Fetches the request identifier from the message.

  int cid() const;
  // Effects: Fetches the client identifier from the message.

  int id() const;
  // Effects : Also fetches the client identifier from the message

  int tsize() const;
  // Effects: Fetches the total size of the proof in the locally committed messa

  int nelem() const;
  
  bool full() const;
  // Effects: Returns true iff "this" is a full reply.

  Digest &request_history_digest() const;
  // Effects : Fetches the request-history digest from the message

  bool verify();
  // Verify if the message is authenticated

  bool verify(int &mcount);
  // Effects: Verifies if the message is authenticated by rep().replica.

  bool match_rh_digest(Digest rh_digest);
  // Effects : Returns true if the request history digest in the reply 
  // matches with rh_digest

  bool match(Local_commit *r);
  // Check if the local commit message r matches with mine

  bool append_proof(struct iovec *p, int len, int psize);
  // Append proof to the message 
 
  Order_request_rep* read_order();
  // Return order request

  int* read_id_list(int reps);
  // return id list

  static bool convert(Message *m1, Local_commit *&m2);
  // Effects: If "m1" has the right size and tag of a "Local_commit", casts
  // "m1" to a "Local_commit" pointer, returns the pointer in "m2" and
  // returns true. Otherwise, it returns false. Convert also trims any
  // surplus storage from "m1" when the conversion is successfull.

private:
  Local_commit_rep &rep() const;
  // Effects: Casts "msg" to a Local_commit_rep&
  
};

inline Local_commit_rep& Local_commit::rep() const { 
  th_assert(ALIGNED(msg), "Improperly aligned pointer");
  return *((Local_commit_rep*)msg); 
}
   
inline View Local_commit::view() const { return rep().v; }
inline Seqno Local_commit::seqno() const { return rep().seq_num; }

inline Request_id Local_commit::request_id() const { return rep().rid; }

inline int Local_commit::cid() const { return rep().cid; }
inline int Local_commit::id() const { return rep().cid; }

inline Digest& Local_commit::request_history_digest() const { return rep().rh_digest; }

inline int Local_commit::tsize() const { return rep().tsize; }

inline int Local_commit::nelem() const { return rep().nelem; }

inline bool Local_commit::match(Local_commit *r) {
  return ((rep().rh_digest == r->rep().rh_digest) && (rep().v == r->rep().v) && (rep().cid == r->rep().cid) && (rep().rid == r->rep().rid));
}

inline bool Local_commit::match_rh_digest(Digest rh_digest) {
  return (rep().rh_digest == rh_digest);
}

inline bool Local_commit::full() const { return true; }

inline Order_request_rep* Local_commit::read_order() {
  return (Order_request_rep *)(contents()+sizeof(Local_commit_rep)+
		      node->auth_size_client());
}

inline int* Local_commit::read_id_list(int reps) {
  int *ptr = (int *)(contents()+sizeof(Local_commit_rep)+node->auth_size_client()
	  +sizeof(Order_request_rep)+reps*node->auth_size_rep());
  return ptr;
}

#endif // _Local_commit_h
