#ifndef _NV_info_h
#define _NV_info_h 1

#include <sys/time.h>
#include "Bitmap.h"
#include "parameters.h"
#include "types.h"
#include "Time.h"
#include "Digest.h"
#include "Pre_prepare_info.h"
#include "Array.h"

class View_info;
class View_change;
class New_view;
class Pre_prepare;
class View_change_ack;
class Status;
class Pre_prepare;
class Prepare;

class NV_info {
  //
  // Holds information concerning a specific new-view message.
  //
public:
  NV_info();
  // Effects: Creates an empty object.
    
  ~NV_info();
  // Effects: Discards any new-view, view-change and view-change ack
  // messages stored in this.
  
  void clear();
  // Effects: Makes this empty -- deletes all contained messages
  // and sets view() == 0.

  View_change *view_change(int id);
  // Effects: If there is any view-change message from replica "id"
  // stored in this returns it. Otherwise, returns 0.

  New_view *new_view(Time **t=0);
  // Effects: If there is a new-view message stored in "this"
  // returns it. Otherwise, returns 0.

  
  View view() const;
  // Effects: Returns the view number of the messages currently
  // stored on this or 0 if this is empty (i.e., if it does not
  // contain any new-view message.)

  bool complete() const;
  // Effects: Returns true iff this contains all the necessary
  // information to move to the new-view.  

  void make_complete(View v);
  // Effects: Mark this as complete in view "v".

  bool add(View_change* vc, bool verified);
  // Requires:  "vc->view() == view()" and "verified == vc->verify()"
  // Effects: If "vc" is one of the messages referenced in the
  // new-view message contained in this and is valid , add "vc" to
  // this and return true. Otherwise, do nothing and return false.

  bool add(New_view* m, View_info* parent);
  // Add new view message 

  
  bool add(View_change_ack* vca);
  // Requires: "vca->view() == view() && vca.verify()"
  // Effects: If there is no view-change corresponding to "vca"
  // in this or referenced by a new-view in this, returns false.
  // Otherwise, it inserts "vca" in this (if its digest matches the
  // view-change's) or deletes vca (otherwise) and returns true.

  void mark_stable(Seqno ls);
  // Effects: Informs "this" that checkpoint sequence number "ls" is
  // stable.

  bool fetch_history_digest(Seqno n, Digest &d);
  // Return history digest if present 

 
 private:
  View v;
  New_view *nv;
  int vc_cur;    // Number of view-change messages already matched with target.
  int vc_target;    // Number of view-change messages already matched with target.
  int vack_count;  // number of view-change messages with 2f+1 vacks

  // Buffer for view changes associated with "nv" and their acks.
  struct VC_info {
    View_change* vc;
    int ack_count;
    Bitmap ack_reps;
    bool req_sum;

    VC_info();
    ~VC_info();
    void clear();
  }; 
  Array<VC_info> vcs;
  Bitmap vcid;

  struct Req_sum {
    View v;
    Digest d; // Digest
    Digest rd; // Request history digest
    int n_proofs; // Number of matches
    Bitmap r_proofs;    // Replicas that sent positive proofs.
    int n_le; 
    int id; // identifier of first replica proposing this
    Req_sum(); 
    ~Req_sum();
  };
  Array<Array<Req_sum> > reqs; // reqs contains an array for each sequence number 
                               // above min an less than or equal to max.
  Array<int> comp_reqs; // For each row index in reqs, contains either -1 
                        // if no request with complete information for that 
                        // index or the column index of the complete Req_sum.  

  Seqno chosen_last_stable;    // Seqno of last stable
  int chosen_last_stable_id;   // Chosen last stable id
  Seqno chosen_max_committed;  // Seqno of max committed locally of all vcs
  int  chosen_committed_id;    // id of the replicas that sent choosen max committed 
  Seqno chosen_max_executed;  // last executed  
  bool done_last_stable;       // last stable is set or not
  bool done_max_committed;     // Max committed is set or not
  Seqno last_complete;         // Last request that has complete information
  
  Seqno min;             // Sequence number of chosen checkpoint
  Seqno max;             // All requests that will propagate to the next view have 
                         // sequence number less than max.
  Seqno base;            // reqs and comp_reqs are indexed by sequence number minus base.
  int n_complete;        // Number of complete entries with seqnos between min and max.

  
  View_info *vi; // Pointer to parent view-info.

  bool is_complete; // True iff this contains all the necessary information for a new-view

  Time nv_sent; // time at which my new-view was last sent

  //
  // Auxiliary methods:

  void summarize();
  // Effects: Summarizes the information in "vc" and combine it with
  // information from other view-change messages.
  
  void summarize_reqs(View_change *vc);
  // Effects: Summarize the request information in "vc" and combine it
  // with information from other view-change messages. 
  
  void choose_ckpt(int index);
  // Requires: "c = sum->ckpts[index]" has n_proofs and n_le greater
  // than or equal to "node->n_f()" 
  // Effects: If "c" has a higher max_stable, than  sum->ckpts[chosen]
  // make it the new chosen checkpoint. Otherwise, do nothing.
  
  void check_complete();
  // Requires: !complete()
  // Effects: Checks if the request corresponding to cur has enough information to 
  // be propagated to the next view. If it does and complete() becomes true, calls
  // make_new_view.

    void make_new_view();
  // Requires: complete() 
  // Effects: Completes the construction of the new-view message.

  bool check_new_view();
  // Effects: Checks if this contains a correct new-view and if so
  // returns true.

};

inline View NV_info::view() const {
  return v;
}

inline bool NV_info::complete() const {
  return is_complete;
}

inline New_view *NV_info::new_view(Time **t=0) {
  if (t && nv)
    *t = &nv_sent;
  return nv;
}

inline  View_change *NV_info::view_change(int id) {
  return vcs[id].vc;
}


#endif
