#include <strings.h>
#include "th_assert.h"
#include "Message_tags.h"
#include "Replica.h"
#include "Order_request.h"
#include "Request.h"
#include "Reply.h"
#include "Req_queue.h"
#include "Req_cache.h"
#include "Request_map.h"
#include "Request_history.h"
#include "Principal.h"
#include "Log_allocator.h"

#define MAX_NON_DET_SIZE 8

Order_request::Order_request(Order_request_rep *cont){
  th_assert(ALIGNED(cont), "Improperly aligned pointer"); 
  msg = cont;
  max_size = -1; // To prevent contents from being deallocated or trimmed
} 

Order_request::Order_request(View view, Seqno s, Request_map &rqueue, Req_history_log &rlog, Req_cache &requests) : 
  Message(Order_request_tag, Max_message_size) {
  rep().v = view;
  rep().seq_num = s;
  int num_reqs= rqueue.unordered_size();
  int bcount  = (num_reqs < max_bsize) ? num_reqs : max_bsize ;
  rep().bsize = bcount; 
  rep().rep_id = node->id(); 
  // Append batched requests information
  char *nexte = contents()+sizeof(Order_request_rep);
  int cur = 0;

  // Fill non-deterministic choices
  char ndet[MAX_NON_DET_SIZE]; 
  int nd_size = replica->max_nd_bytes();
  if (bcount > 0) {
    // TO DO : check if we have enough size for the ND values
    replica->compute_non_det(s,ndet,&nd_size);
    rep().non_det_size = nd_size;
  }
  else {
     th_assert(0,"Primary : there are no buffered requests\n");
  }

  for (Request *req = rqueue.unordered_remove(); req != 0 && cur < bcount; req = rqueue.unordered_remove()) {
   
    // Copy request id information
    int cid = req->client_id();
    Request_id rid = req->request_id();
    Digest &d = req->digest();

    memcpy(nexte,(char*)&cid,sizeof(int));
    nexte += sizeof(int);
    memcpy(nexte,(char *)&rid,sizeof(Request_id));
    nexte += sizeof(Request_id);
    memcpy(nexte,(char *)&d,sizeof(Digest)); 
    nexte += sizeof(Digest);

    // Update the request history log
    // TO DO : We are paying cost for computing request history digest for each 
    // request. We can batch this and reduce the cost ??
    rlog.add_request(s+cur, cid, rid, d, ndet, nd_size, s+bcount-1);
    cur++;

    // Update the request map of primary
    rqueue.order_request(req,cid,rid,d,view,s);  

    // Add request to the request cache
    requests.add(cid,rid,this);

    // Check for commit flag 
    if (req->is_set_commit()) {
      rep().extra |= 4;
    }
  }

  // copy the non det
  if (nd_size) {
    memcpy(nexte, (char *)ndet, nd_size);
    nexte += nd_size;
  }
  assert(cur == bcount);
  
  // Copy request history digest
  rlog.comp_history_digest(s+bcount-1,&(request_history_digest()));
  
  // Sign the order_request message
  int old_size =  sizeof(Order_request_rep)+
    bcount*(sizeof(int)+sizeof(Request_id)+sizeof(Digest))+nd_size;
  
  set_size(old_size);
 
}

    
Digest* Order_request::digest(Seqno s) {
  // Always call is_present before calling this function
  Seqno base = seqno();
  
  return  (Digest*)(contents()+sizeof(Order_request_rep)+
    (s-base)*(sizeof(int)+sizeof(Request_id)+sizeof(Digest))+
    sizeof(int)+sizeof(Request_id));
}

Digest& Order_request::req_digest(Seqno s) {
  th_assert(s <= seqno(), "Invalid argument range \n");
  Digest *d = digest(s); // Digest of the base seq no
  return *d;
}

int* Order_request::cid(Seqno s) {
  // Always call is_present before calling this function
  Seqno base = seqno();

   return  (int*)(contents()+sizeof(Order_request_rep)+
    (s-base)*(sizeof(int)+sizeof(Request_id)+sizeof(Digest)));
}

Request_id* Order_request::rid(Seqno s) {
  // Always call is_present before calling this function
  Seqno base = seqno();
  
  return  (Request_id*)(contents()+sizeof(Order_request_rep)+
    (s-base)*(sizeof(int)+sizeof(Request_id)+sizeof(Digest))+
    sizeof(int));
}


char* Order_request::choices(int &len) {
  // Return the non-det 
  len = rep().non_det_size;
  return (char *)(contents()+sizeof(Order_request_rep)+
		  (bsize())*(sizeof(int)+sizeof(Request_id)+sizeof(Digest)));
}
  
Order_request* Order_request::copy(int id) const {
  Order_request* ret = (Order_request*)new Message(msg->size);
  memcpy(ret->msg, msg, msg->size);
  ret->rep().rep_id = id;
  return ret;
}

Request_id Order_request::request_id(int id) {
	
  for (int i=seqno();i<seqno()+bsize();i++) {
    int *c = cid(i);
    if (id == *c) {
      Request_id *req_id= rid(i);
      return *req_id;
    }
  }

  return Long_max;
}
 
// This is called by the primary to send order requests to all clients 
void Order_request::authenticate_send_clients() {
  int old_size = size();
  
  // Authenticate the message for each client and send it
  // PKI : No need to sign again
  for (int i=seqno();i<seqno()+bsize();i++) {
    int *c = cid(i);
   
    // ZYZZYVA : Cache client MACs as well we do not need to compute this 
    // on retransmission
    set_size(old_size+MAC_size);

#ifndef USE_PKEY 
    Principal *cp = node->i_to_p(*c);
    cp->gen_mac_out(contents(), sizeof(Order_request_rep), contents()+old_size);
#endif
    old_size += MAC_size;
    node->send(this, *c);
  }

  // Trim the message before storing in the primary log
  trim();
}
 
// This is called by backups to generate authenticator 
// before sending the OR message to the client or backup replicas
// This should not be used by the primary.

void Order_request::my_auth() {
  int bcount= bsize();
  int nd_size = non_det_size();
  th_assert(node->id() != node->primary(), "Primary using incorrect function");
  
#ifndef USE_PKEY 
  // Sign the order_request message
  int size_wo_auth =  ALIGNED_SIZE(sizeof(Order_request_rep)+
				   bcount*(sizeof(int)+sizeof(Request_id)+sizeof(Digest))+nd_size);
  int old_size =  size_wo_auth+node->auth_size_rep();
  
  // Check if there is enough space
  th_assert((unsigned) node->auth_size() < (unsigned) (msize()-old_size),
	    "Insufficient memory");

  // First set the size
  set_size(old_size+node->auth_size());   
  trim(); // We trim the received OR request message after authorizing the request

  // First copy primary authenticator to the end
  memcpy(contents()+old_size,contents()+size_wo_auth,node->auth_size());

  // Change the replica id to my id
  change_replica_id(node->id());

  // Reset the size of the msg to the old_size so that when this message 
  // is sent by the backup, primary authenticator is not included
  set_size(old_size);

  // Generate the authenticator in the place of primary authenticator
  node->gen_auth_out(contents(), sizeof(Order_request_rep), contents()+size_wo_auth);


#else 
  int size_wo_auth = ALIGNED_SIZE(sizeof(Order_request_rep)+bcount*(sizeof(int)+sizeof(Request_id)+sizeof(Digest))+nd_size);
  int old_size =  size_wo_auth+node->sig_size();
  
  // Check if there is enough space
  th_assert((unsigned) node->sig_size() < (unsigned) (msize()-old_size),
	    "Insufficient memory");

  // First set the size
  set_size(old_size+node->sig_size());   
  trim(); // We trim the received OR request message after authorizing the request

  // First copy primary authenticator to the end
  memcpy(contents()+old_size,contents()+size_wo_auth,node->sig_size());

  // Reset the size of the msg to the old_size so that when message 
  // is sent by the backup primary authenticator is not included
  set_size(old_size);

  // Generate the authenticator in the place of primary authenticator
  node->gen_signature(contents(), sizeof(Order_request_rep), contents()+size_wo_auth); 
  

#endif
  
}

// This is called by the primary to authenticate other replicas
void Order_request::authenticate() {
  //int bcount = bsize();
  //int nd_size = rep().non_det_size;
  int old_size = size();
  //fprintf(stderr,"Authenticate size : %d....\n", old_size); 
  //int old_size=sizeof(Order_request_rep)+bcount*(sizeof(int)+sizeof(Request_id)+sizeof(Digest))+nd_size; 

  th_assert((unsigned) node->auth_size() < (unsigned) (msize()-old_size),
	    "Insufficient memory");

#ifndef USE_PKEY 
  // MAC optimzation is not set
  // generate authenticator with MACs for other replicas

  // WARNING : FIX THIS 
  // There can be man in the middle attack as we are not including the 
  // rid, digest, ND contents in the signature
  // If we change this we may have to change Reply::authenticate_copy_order
  // to include all the above fields in the reply
  set_size(old_size+node->auth_size());   
  node->gen_auth_out(contents(), sizeof(Order_request_rep), contents()+old_size);
  //fprintf(stderr,"Sending order request size : %d....\n", size()); 
  
#else 
  set_size(old_size+node->sig_size());
  node->gen_signature(contents(), sizeof(Order_request_rep), contents()+old_size); 
#endif

  //trim();
}

void Order_request::re_authenticate(Principal *p, bool mac) {
  // TO DO : Re-authenticate a principal using the new key
  if (p) {
    int old_size = size();
    
    if (mac) {
      th_assert((unsigned) MAC_size < (unsigned )(msize()-old_size),
	    "Insufficient memory");
      p->gen_mac_out(contents(),sizeof(Order_request_rep),contents()+old_size);
      set_size(old_size+MAC_size);
    }
    else {
      th_assert((unsigned) node->auth_size() < (unsigned) (msize()-old_size),
		"Insufficient memory");
      node->gen_auth_out(contents(), sizeof(Order_request_rep), contents()+sizeof(Order_request_rep));
      set_size(old_size+node->auth_size()); 
    }
    trim();
  }
}

bool Order_request::verify_client() {
#ifndef USE_PKEY
  bool is_mac_opt  = rep().extra;
  if (size()-(int)sizeof(Order_request_rep) < MAC_size)
    return false;

  // Check reply
  if (!is_mac_opt) {
    if (size()-(int)sizeof(Order_request_rep)-MAC_size < node->auth_size())
        return false;
  }

  //fprintf(stderr,"Received order request size : %d....\n", size());
  // Check MAC
  INCR_OP(reply_auth_ver);
  START_CC(reply_auth_ver_cycles);
  int offset = size()-MAC_size;
  char *dest = contents()+offset;
  Principal *rp = node->i_to_p(id());
  bool ret = rp->verify_mac_in(contents(), sizeof(Order_request_rep), dest); 

#else
 // PKI 
#endif

  //fprintf(stderr,"Returning  verify %d ....\n",ret); 
  return ret;
}
 
bool Order_request::verify() {

  if (node->id() >= node->n()) {
    return verify_client();
  }

  // Check sizes
#ifndef USE_PKEY
  bool is_mac_opt  = rep().extra;
  if (size()-(int)sizeof(Order_request_rep) < MAC_size) 
    return false;

  // Check reply
  if (!is_mac_opt) {
    if (size()-(int)sizeof(Order_request_rep)-MAC_size < node->auth_size()) 
        return false;
  }

  //fprintf(stderr,"Received order request size : %d....\n", size()); 
  // Check MAC 
  INCR_OP(reply_auth_ver);
  START_CC(reply_auth_ver_cycles);
  int bcount = bsize();
  int nd_size = rep().non_det_size;  
  int offset = ALIGNED_SIZE(sizeof(Order_request_rep)+
			    bcount*(sizeof(int)+sizeof(Request_id)+sizeof(Digest))+nd_size);
  //fprintf(stderr,"Entered verify size : %d....\n", offset); 
  char *dest = contents()+offset;
  bool ret = node->verify_auth_in(id(), contents(), sizeof(Order_request_rep), dest); 
  STOP_CC(reply_auth_ver_cycles);

#else
  int bcount = bsize();
  int nd_size = rep().non_det_size;  
  int offset = ALIGNED_SIZE(sizeof(Order_request_rep) + 
			    bcount*(sizeof(int) + sizeof(Request_id) + sizeof(Digest)) 
			    + nd_size);

  if (size()-(int)sizeof(Order_request_rep) < node->sig_size(id())) 
      return false; 
 
  ret = node->i_to_p(id())->verify_signature(contents(), sizeof(Order_request_rep), offset); 
#endif

  //fprintf(stderr,"Returning  verify %d ....\n",ret); 
  return ret;
}

bool Order_request::convert(Message *m1, Order_request *&m2) {
  if (!m1->has_tag(Order_request_tag, sizeof(Order_request_rep)))
    return false;
  
  //  m1->trim(); We trim the OR message after authenticating the message
  m2 = (Order_request*)m1;
  return true;
}

