#ifndef _Order_request_h
#define _Order_request_h 1

#include "types.h"
#include "Message.h"
#include "Digest.h"

class Principal;
class Rep_info;
class Req_cache;
class Request_map;
class Req_history_log;

// 
// Order_request messages have the following format.
//
// Order_request_rep.extra = 4 if commit opt is on
// 

struct Order_request_rep : public Message_rep {
  View v;                // Current view
  int seq_num;           // Sequence number
  Digest rh_digest;      // request history digest
  int rep_id;            // id of replica sending the reply
  int bsize;             // batch size
  int non_det_size;      // Non-deterministic choices
  
  // 1. Followed by client_id[], rid[], non-det[] for the batch of requests

  // 2. Followed by a MAC or an authenticator. It includes 
  //    authenticator in an un-optimized version. 
  //    TO DO : In the optimized 
  //    version a replica just inlcudes the MAC for the client 
  //    but not for other replicas 
 
};

class Order_request: public Message {
  // 
  // Order_request messages
  //
public:
  Order_request() : Message() {}

  Order_request(View view, Seqno s, Request_map &rqueue, Req_history_log &rhlog, Req_cache &reqc);

  Order_request(View view, int client_id, Request_id req, int seq_num, Digest rh_digest, int replica, bool is_mac, int auth_size, Principal *p);
  // Effects: Creates an order-request message for a given request (rid) with request-history digest 
  // sequence number
  // authentication. The method store_reply and authenticate should
  // be used to finish message construction.

  Order_request(Order_request_rep *contents);
  // Requires: "contents" contains a valid Order_request_rep.
  // Effects: Creates a message from "contents". No copy is made of
  // "contents" and the storage associated with "contents" is not
  // deallocated if the message is later deleted. Useful to create
  // messages from reps contained in other messages.


  Order_request* copy(int id) const;
  // Effects: Creates a new object with the same state as this but
  // with replica identifier "id"
  
  char* choices(int &len);
  // Effects: Returns a buffer that can be filled with non-deterministic choices

  void authenticate();
  // Effects: Terminates the construction of a order_request message by
  // setting the length of the reply to "act_len", appending a MAC/authenticator,
  // and trimming any surplus storage.

  void authenticate_send_clients();
  // Authenticate and send the order request messages with MACs for all replicas 
  // to all clients in the batch of requets

  void re_authenticate(Principal *p, bool mac);
  // Effects: Recomputes the authenticator in the reply using the most
  // recent key.

  View view() const;
  // Effects: Fetches the view from the message

  int rep_id() const;
  // Return the replica sending the request

  int id() const;
  // Effects: Fetches the reply identifier from the message.

  Digest &request_history_digest() const;
  // Effects : Fetches the request-history digest from the message

  bool is_mac() const;
  // Effects: Returns true iff "this" is a full reply.

  bool is_present(Seqno s);
  // Is request s batched in this order-request message

  int bsize();
  // Return bsize

  int non_det_size();
  // Return non determinant variables size  

  char* request_contents() ; 
  // Effects: Return the request contents

  char* auth_contents();
  // Effects: Return the authenticator contents

  bool verify();
  // Effects: Verifies if the message is authenticated by rep().replica.

  bool verify_client();
  // Effects: Client verifies if the message is authenticated by rep().replica.
  
  bool match(Order_request *r);
  // Effects: Returns true if the replies match.
  
  static bool convert(Message *m, Order_request *&r);
  // Converts the incomming message to order-request message
  
  bool commit_opt();
  // Return true if commit opt bit is set in the client request
  
  Seqno seqno();
  // Return the base sequence number 
  
  void my_auth();
  // AUthorize the received OR request message

  void change_replica_id(int rid);
  // Channge the repica id in the Order_Request_rep

  int* cid(Seqno s);
  // Returns client id of the request with sequence s

  Digest* digest(Seqno s);
  // Returns the digest of the request with seq no s

  Digest& req_digest(Seqno s);
  // Returns the digest of request with seqno s in the batch.

  Request_id* rid(Seqno s);
  // Returns the rid of the request with seq no s

  Request_id request_id();
  // DO NOT USE THIS FUNCTION 

  Request_id Order_request::request_id(int client_id); 
  // Return request id of the client cid in the batch

private:
  Order_request_rep &rep() const;

  // Effects: Casts "msg" to a Reply_rep&
  // MAX batch size 
  static const int max_bsize = 100;
};

inline bool Order_request::is_present(Seqno s) {
  return (s <= (rep().seq_num+rep().bsize-1));
}

inline Order_request_rep& Order_request::rep() const { 
  th_assert(ALIGNED(msg), "Improperly aligned pointer");
  return *((Order_request_rep*)msg); 
}

inline Request_id  Order_request::request_id() { return 0;}

inline void Order_request::change_replica_id(int rid) {rep().rep_id = rid;}

inline int  Order_request::bsize() { return rep().bsize;} 

inline int Order_request::non_det_size() { return rep().non_det_size;}

inline Seqno Order_request::seqno() { return rep().seq_num;}

inline View Order_request::view() const { return rep().v; }

inline int Order_request::rep_id() const { return rep().rep_id; }

inline bool Order_request::is_mac() const { return rep().extra == 1; }

inline int Order_request::id() const { return rep().rep_id; }

inline Digest& Order_request::request_history_digest() const { return rep().rh_digest; }

inline char* Order_request::request_contents() { 
  return contents()+sizeof(Order_request_rep);
}
  
inline bool Order_request::match(Order_request *r) {
  //fprintf(stderr,"Cur: View : %llu Seq num : %d \n", r->view(), r->rep().seq_num);
  //fprintf(stderr,"Org: View : %llu Seq num : %d \n", view(),rep().seq_num );
  // TO DO: Should also check non-deterministic choices, batch size
  return ((rep().rh_digest == r->rep().rh_digest) && (view() == r->view())
          && (rep().seq_num == r->rep().seq_num));
}

inline char *Order_request::auth_contents() {
    int bcount= bsize();
  int nd_size = non_det_size();
  int old_size =  sizeof(Order_request_rep)+
    bcount*(sizeof(int)+sizeof(Request_id)+sizeof(Digest))+nd_size;
  return contents()+old_size;
}

inline bool Order_request::commit_opt() {
  return rep().extra & 4;
}
  
  
#endif // _Order_request_h
