#include <strings.h>
#include "th_assert.h"
#include "Message_tags.h"
#include "Reply.h"
#include "Node.h"
#include "Principal.h"

#include "Statistics.h"

Reply::Reply(View view, Request_id req, Seqno snum, Seqno last_committed, Digest &rh_d, int replica) : 
  Message(Reply_tag, Max_message_size) {
    rep().v = view; 
    rep().rid = req;
    rep().seq_num = snum;
    rep().rep_id = replica;
    rep().rh_digest = rh_d;
    rep().reply_size = -2;    
    rep().last_committed = last_committed;

}


Reply::Reply(Reply_rep *r) : Message(r) {}


Reply::Reply(View view, Request_id req, Seqno snum, Seqno last_committed, int replica, Digest &rh_d, Digest &rep_d, 
	     Principal *p) : Message(Reply_tag, Max_message_size) {
    rep().v = view; 
    rep().rid = req;
    rep().seq_num = snum;
    rep().rep_id = replica;
    rep().reply_size = -1;
    rep().rh_digest = rh_d;
    rep().rep_digest = rep_d;
    rep().last_committed = last_committed;
}


Reply* Reply::copy(int id) const {
  Reply* ret = (Reply*)new Message(msg->size);
  memcpy(ret->msg, msg, msg->size);
  ret->rep().rep_id = id;
  return ret;
}


char *Reply::store_reply(int &max_len) {
  int osize = order() ? sizeof(Order_request_rep)+node->auth_size() : 0; 
  max_len = msize()-sizeof(Reply_rep)-osize-MAC_size;
  return contents()+sizeof(Reply_rep)+osize;
}


void Reply::authenticate(Principal *p, int act_len) {
  th_assert((unsigned)act_len <= msize()-sizeof(Reply_rep)-MAC_size,
	    "Invalid reply size");

  int osize = order() ? sizeof(Order_request_rep)+node->auth_size() : 0;

  rep().reply_size = act_len;
  rep().rep_digest = Digest(contents()+sizeof(Reply_rep)+osize, act_len);
  int old_size = sizeof(Reply_rep)+act_len+osize;
  set_size(old_size+MAC_size);

  INCR_OP(reply_auth);
  START_CC(reply_auth_cycles);
  p->gen_mac_out(contents(), sizeof(Reply_rep)+osize, contents()+old_size);
  STOP_CC(reply_auth_cycles);

  trim();
}

void Reply::comp_digest(Digest &d, int act_len) {
  int osize = order() ? sizeof(Order_request_rep)+node->auth_size() : 0;
  d = Digest(contents()+sizeof(Reply_rep)+osize,act_len);
}

void Reply::re_authenticate(Principal *p) {
  int rs = rep().reply_size;

  int old_size = (rs > 0) ? sizeof(Reply_rep)+rs : sizeof(Reply_rep);
  int osize = order() ? sizeof(Order_request_rep)+node->auth_size() : 0;
  old_size += osize;

  set_size(old_size+MAC_size);

  INCR_OP(reply_auth);
  START_CC(reply_auth_cycles);
  p->gen_mac_out(contents(), sizeof(Reply_rep)+osize, contents()+old_size);
  STOP_CC(reply_auth_cycles);
  
  trim();
}

void Reply::authenticate_copy_order(Order_request *oreq) {
  set_order();// set the order bit
  char *src = contents()+sizeof(Reply_rep);

  // Copy the OR header (No Need to copy the request information in the batch)
  memcpy(src, oreq->contents(), sizeof(Order_request_rep));
  
  // Copy the authenticator
  memcpy(src+sizeof(Order_request_rep), oreq->auth_contents(), node->auth_size_rep());

}

bool Reply::verify() {

  // Replies must be sent by replicas.
  if (!node->is_replica(id())) 
    return false;

  // Check sizes
  int rep_size = (full()) ? rep().reply_size : 0;
  int osize = (order()) ? sizeof(Order_request_rep)+node->auth_size(0) : 0;

  if (size()-(int)sizeof(Reply_rep)-rep_size - osize < MAC_size) 
    return false;

  // Check reply
  if (full()) {
    Digest d(contents()+sizeof(Reply_rep)+osize, rep_size);
    if (d != rep().rep_digest)
      return false;
  }

  // Check signature.
  Principal *rp = node->i_to_p(rep().rep_id);
  int size_wo_MAC = sizeof(Reply_rep)+rep_size+osize;
  
  INCR_OP(reply_auth_ver);
  START_CC(reply_auth_ver_cycles);

  bool ret = rp->verify_mac_in(contents(), sizeof(Reply_rep)+osize, contents()+size_wo_MAC);

  STOP_CC(reply_auth_ver_cycles);

  return ret;
}


bool Reply::convert(Message *m1, Reply *&m2) {
  if (!m1->has_tag(Reply_tag, sizeof(Reply_rep)))
    return false;
  
  m1->trim();
  m2 = (Reply*)m1;
  return true;
}
