#ifndef _Reply_h
#define _Reply_h 1

#include "types.h"
#include "Message.h"
#include "Digest.h"
#include "Order_request.h"

class Principal;
class Rep_info;
class Node;

#define INCLUDE_ORDER_MSG 2
#define INCLUDE_LOCAL_COMMIT_NUM 4

// 
// Reply messages have the following format.
//
struct Reply_rep : public Message_rep {
  View v;                // Current view
  Request_id rid;        // unique request identifier
  Seqno seq_num;         // Sequence number of the reply
                         // (Primary optimization : we need Seq num as rid is not present in the 
                         // order request sent by the primary to the client)
  Seqno last_committed;  // Read only optimization: requires this to return result in the 
                         // presence of failures with just 2f+1 matching replies
  Digest rh_digest;      // request history digest
  Digest rep_digest;     // reply digest
  int rep_id;            // id of replica sending the reply
  int reply_size;        // if negative, reply is not full.
  
  // 1. Followed by order request sent by the primary 
  //    of size Order_request_rep + authenticator size

  // 2. Followed by the application reply that is "reply_size" bytes 
  //    long if it is from an execution replica 
  //     -- Bandwidth optizization : 
  //      Replies can empty or full. Only one execution 
  //      replica (chosen randomly by the client) sends the full reply
  //      others just send the digest

  
  // 3. Followed by MAC authenticating the reply to the client. 
  
};

class Reply : public Message {
  // 
  // Reply messages
  //
public:
  Reply() : Message() {}

  Reply(Reply_rep *r);

  Reply(View view, Request_id req, Seqno s, Seqno last_committed, Digest &rh_d,int replica);
  // Effects: Creates a new (full) Reply message with an empty reply and no
  // authentication. The method store_reply and authenticate should
  // be used to finish message construction.

  Reply(View view, Request_id req, Seqno s, Seqno last_committed, int replica, Digest &rh_d, Digest &rep_d, 
	Principal *clientp);
  // Return reply to "clientp" with history digest "rh_d", reply digest, "rep_d"
  // request rid "req"

  Reply* copy(int id) const;
  // Effects: Creates a new object with the same state as this but
  // with replica identifier "id"

  char *store_reply(int &max_len);
  // Effects: Returns a pointer to the location within the message
  // where the reply should be stored and sets "max_len" to the number of
  // bytes available to store the reply. The caller can copy any reply
  // with length less than "max_len" into the returned buffer. 

  void authenticate(Principal *p, int act_len);
  // Effects: Terminates the construction of a reply message by
  // setting the length of the reply to "act_len", appending a MAC,
  // and trimming any surplus storage.
  
  void authenticate_copy_order(Order_request *oreq);
  // Copy and authenticate the order request message in the reply
  // IMPORTANT: OR request message should be copied before the reply
  // is authenticated
  
  void re_authenticate(Principal *p);
  // Effects: Recomputes the authenticator in the reply using the most
  // recent key.

  View view() const;
  // Effects: Fetches the view from the message

  Seqno seqno() const;
  // Returns the seq no of the request 

  Request_id request_id() const;
  // Effects: Fetches the request identifier from the message.

  int rep_id() const;
  // Effects: Fetches the reply identifier from the message.

  int id() const;
  // Effects: Fetches the reply identifier from the message.
  
  Digest &digest() const;
  // Effects : Fetches the reply digest from the message

  Digest &request_history_digest() const;
  // Effects : Fetches the request-history digest from the message

  Digest &reply_digest() const;
  // Effects : Fetches the reply digest from the message
  
  char *reply(int &len);
  // Effects: Returns a pointer to the reply and sets len to the
  // reply size.
  
  int& reply_size();
  // Return the reply size

  bool full() const;
  // Effects: Returns true iff "this" is a full reply.

  bool order() const;
  // Return true if the order_request message is included

  void set_order();
  // Sets the order_request include bit in the "extra" field of the reply_rep structure

  Order_request_rep* read_order();
  // Read the order request message from the reply

  char* read_order_auth();

  int order_size();
  // returns the size of the order message
  
  bool check_locally_committed();
  // Check if current reply is locally committed
    
  bool verify();
  // Effects: Verifies if the message is authenticated by rep().replica.

  bool match(Reply *r);
  // Effects: Returns true if the replies match.

  bool match_rh_digest(Digest rh_digest);
  // Effects : Returns true if the request history digest in the reply 
  // matches with rh_digest

  void comp_digest(Digest &d, int act_len);
  // Effects : Compute digest

  static bool convert(Message *m1, Reply *&m2);
  // Effects: If "m1" has the right size and tag of a "Reply", casts
  // "m1" to a "Reply" pointer, returns the pointer in "m2" and
  // returns true. Otherwise, it returns false. Convert also trims any
  // surplus storage from "m1" when the conversion is successfull.

private:
  Reply_rep &rep() const;
  // Effects: Casts "msg" to a Reply_rep&
  
  friend class Rep_info;
};

inline Reply_rep& Reply::rep() const { 
  th_assert(ALIGNED(msg), "Improperly aligned pointer");
  return *((Reply_rep*)msg); 
}
   
inline View Reply::view() const { return rep().v; }

inline Request_id Reply::request_id() const { return rep().rid; }

inline int Reply::rep_id() const { return rep().rep_id; }

inline Seqno Reply::seqno() const { return rep().seq_num; }

inline int Reply::id() const { return rep().rep_id; }

inline Digest& Reply::reply_digest() const { return rep().rep_digest; }

inline Digest& Reply::digest() const { return rep().rep_digest; }

inline Digest& Reply::request_history_digest() const { return rep().rh_digest; }

inline int& Reply::reply_size()  {return rep().reply_size;} 

inline char* Reply::reply(int &len) { 
  len = rep().reply_size;
  return contents()+sizeof(Reply_rep);
}
  
inline bool Reply::full() const { return rep().reply_size >= 0; }

inline bool Reply::order() const { return rep().extra && INCLUDE_ORDER_MSG; }

inline void Reply::set_order() { rep().extra |= INCLUDE_ORDER_MSG;}

inline Order_request_rep* Reply::read_order() {
  if (!order()) return NULL;

  return ((Order_request_rep *) (contents()+sizeof(Reply_rep)));
}

inline char *Reply::read_order_auth() {
  if (!order()) return NULL;

  return ((char *) contents()+sizeof(Reply_rep)+sizeof(Order_request_rep));
}

inline int Reply::order_size() { return sizeof(Order_request_rep);}



inline bool Reply::match(Reply *r) {
  return ((request_history_digest() == r->request_history_digest()) && 
         ((r->reply_size() < 0) || (reply_size() < 0) || (digest() == r->digest())) &&
         (rep().v == r->rep().v) &&
         (rep().rid == r->rep().rid) && 
	 (rep().seq_num == r->rep().seq_num)); 
}

inline bool Reply::match_rh_digest(Digest rh_digest) {
  return (rep().rh_digest == rh_digest);
}

inline bool Reply::check_locally_committed() {
  return (rep().last_committed >= rep().seq_num);
}

#endif // _Reply_h
