#include <string.h>

#include "th_assert.h"
#include "Message_tags.h"
#include "View_change.h"
#include "Replica.h"
#include "Principal.h"
 
View_change::View_change(View v, Seqno ls, Seqno lc, Seqno le, int id) 
  : Message(View_change_tag, Max_message_size) {
  rep().v = v;
  rep().ls = ls;
  rep().lc = lc;
  rep().le = le;
  rep().id = id;

  // Initialize req and checkpoint info
  rep().chkpt.lv = -1;
  for (int i=ls+1; i <= le; i++) {
    req_info()[i-ls-1].lv = -1;
  } 

}

void View_change::add_checkpoint(Seqno n, Digest &d, View lv) {
  th_assert(last_stable()==n, "Invalid argument");

  rep().chkpt.rd = d;
  rep().chkpt.n = n;
  rep().chkpt.lv = lv;
  fprintf(stderr, "Adding checkpoint in view_change: ");
  d.print();
}
 

void  View_change::add_request(Seqno n, View lv, Digest &d, Digest &rd) {
  th_assert(last_stable() < n <= last_stable()+max_out, "Invalid argument");
  th_assert(lv < view(), "Invalid argument");
  
  int index = n-last_stable()-1;
  
  Req_info &ri = req_info()[index];
  ri.rd = rd;
  ri.d = d;
  ri.n = n;
  ri.lv = lv;
}

View View_change::request(Seqno n, Digest &d, Digest &rd) {
  th_assert(n > last_stable(), "Invalid argument");
 
  int index = n-last_stable()-1;
   
  Req_info &ri = req_info()[index];
  d = ri.d;
  rd = ri.rd;
  return ri.lv; 
}

bool View_change::req(Seqno n, Digest &d, Digest &rd) {
  th_assert(n >= last_stable(), "Invalid argument");
 
  if (n == last_stable()) {
     return ckpt(rd,n);
  }
  
  int index = n-last_stable()-1;
   
  Req_info &ri = req_info()[index];
  d = ri.d;
  rd = ri.rd;
  return true; 
}


void View_change::authenticate(Principal *p) {
  th_assert(last_stable() >= 0 && last_stable()%checkpoint_interval == 0, "Invalid state");

  int old_size = sizeof(View_change_rep)+sizeof(Req_info)*(num_reqs());

  // Compute authenticator and update size.
  set_size(old_size+node->auth_size());

  rep().d.zero(); 
  rep().d = Digest(contents(),old_size);

  node->gen_auth_out(contents(), old_size, contents()+old_size);
}


bool View_change::verify() {
  int nreqs = num_reqs();
  if (!node->is_replica(id()) || nreqs < 0 || nreqs > max_out || view() <= 0) 
    return false;

  if (last_stable() < 0 || last_stable()%checkpoint_interval != 0) 
    return false;
  
  if (last_stable() > last_committed() || last_stable() > last_executed() ||
      last_committed() > last_executed())
    return false;

  // Check sizes
  int old_size = sizeof(View_change_rep)+sizeof(Req_info)*nreqs;

  if (size()-old_size < node->auth_size(id()))
    return false;
  
  // Check consistency of request information.
  for (int i=0; i < nreqs; i++) {
    Req_info &ri = req_info()[i];
    if (ri.lv >= view())
      return false;
  }

  // Check digest of message.
  if (!verify_digest()) {
    return false;
  }
      
  // Check authenticator.
  if (!node->verify_auth_in(id(), contents(), old_size, contents()+old_size))
    return false;

  return true;
}


bool  View_change::verify_digest() {
  Digest d =  digest(); // save old digest
  digest().zero();  // zero digest

  bool verified = 
    (d == Digest(contents(), sizeof(View_change_rep)+sizeof(Req_info)*(last_executed()-last_stable())));
  
  digest() = d; // restore digest
  return verified;
}


bool View_change::convert(Message *m1, View_change  *&m2) {
  if (!m1->has_tag(View_change_tag, sizeof(View_change_rep)))
    return false;

  m1->trim();
  m2 = (View_change*)m1;
  return true;
}
