#ifndef _View_change_h
#define _View_change_h 1

#include "parameters.h"
#include "types.h"
#include "bits.h"
#include "Digest.h"
#include "Principal.h"
#include "Message.h"


//
// Req_infos describe requests that are ordered and/or committed locally.
//  


struct Req_info {
  Digest d;  // Request digest of a request 
  Digest rd; // request history digest 
  Seqno n;   // Sequence number of the request
  View lv;   // View in which this request was last ordered
};


// 
// View_change messages have the following format:
//
struct View_change_rep : public Message_rep {
  View v;   // sending replica's new view
  Seqno ls; // sequence number of last checkpoint known to be stable
  Seqno lc; // last committed request seqno
  Seqno le; // last request that is executed tentatively
  int id; // sending replica's id 

  Req_info chkpt; // Stable checkpoint information

  // digest of the entire message (except authenticator) with d zeroed.
  Digest d;
 
  /* 
     Followed by:
     Req_info req_info[le-ls];
     
     Contains all the requests from lc to lo.

     // This is followed by an authenticator from principal id. 
   */
};


class View_change : public Message {
  // 
  //  View_change messages
  //
public:
  View_change(View v, Seqno ls, Seqno lc, Seqno le, int id);
  // Effects: Creates a new (unauthenticated) View_change message for
  // replica "id" in view "v". The message states that "ls" is the
  // sequence number of the last checkpoint known to be stable but the
  // message has an empty set of requests and checkpoints.

  void add_checkpoint(Seqno n, Digest &rd, View lv);
  // Adds the stable checkpoint

  void add_request(Seqno n, View lv, Digest &d, Digest &rd);
  // Requires: "last_stable() < n <= last_stable()+max_out".
  // Effects: Sets the Req_info for the request with sequence number
  // "n" to "{lv, v, d}" and records whether the request is prepared.

  int id() const;
  // Effects: Fetches the identifier of the replica from the message.

  View view() const;
  // Effects: Returns the view in the message.

  Digest& digest();
  // Effects: Returns the digest of this message (excluding the
  // authenticator).

  Seqno last_stable() const;
  // Effects: Returns the sequence number of the last stable
  // checkpoint.

  Seqno last_committed() const;
  // Effects: Returns the sequence number of the last stable
  // checkpoint.

  Seqno last_executed() const;
  // Effects: Returns the sequence number of the last ordered request without hole
  // checkpoint.
  
  Seqno max_seqno() const;
  // Effects: Returns the maximum sequence number refered to in this.

  int num_reqs() const;
  // Number of requests in Req_info 

  bool ckpt(Digest &rd, Seqno &n);
  // Effects: If this contains some checkpoint digest, returns true
  // and sets "d" to the digest of the checkpoint with the highest
  // sequence number "n" int this.
  
  bool req(Seqno n, Digest &d, Digest &rd);
  View request(Seqno n, Digest &d, Digest &rd);
  // Requires: n > last_stable()
  // Effects: Returns the view and sets "d" and "rd" to the digest associated with
  // the request and request history digest for the request with sequence number 
  // "n" in the message.

  void authenticate(Principal *p=0);
  // Effects: Recomputes the authenticator in the message using the
  // most recent keys. If "p" is not null, may only update "p"'s
  // entry in the authenticator.
  
  bool verify();
  // Effects: Verifies if the message is syntactically correct and
  // authenticated by the principal rep().id.

  bool verify_digest();
  // Effects: Returns true iff digest() is correct.

  static bool convert(Message *m1, View_change *&m2);
  // Effects: If "m1" has the right size and tag of a "View_change",
  // casts "m1" to a "View_change" pointer, returns the pointer in
  // "m2" and returns true. Otherwise, it returns false.
  // If the conversion is successful it trims excess allocation.

private:
  View_change_rep &rep() const;
  // Effects: Casts "msg" to a View_change_rep&
  
  Req_info *req_info();
  // Effects: Returns a pointer to the prep_info array.
    
};


inline View_change_rep &View_change::rep() const { 
  th_assert(ALIGNED(msg), "Improperly aligned pointer");
  return *((View_change_rep*)msg); 
}
  
inline Req_info *View_change::req_info() { 
  Req_info *ret = (Req_info *)(contents()+sizeof(View_change_rep));
  return ret; 
}

inline int View_change::id() const { return rep().id; }

inline  View View_change::view() const { return rep().v; }

inline  Digest& View_change::digest() { 
  return rep().d;
}

inline  Seqno View_change::last_stable() const { return rep().ls; }
inline  Seqno View_change::last_committed() const { return rep().lc; }
inline  Seqno View_change::last_executed() const { return rep().le; }
inline  int View_change::num_reqs() const {return last_executed()-last_stable();}

inline Seqno View_change::max_seqno() const { return rep().le; }

inline  bool View_change::ckpt(Digest& rd, Seqno &n) { 
  if (rep().lc >= 0) {
    rd = rep().chkpt.rd;
    n = rep().chkpt.n;
   
    //fprintf(stderr,"Returning digest by id : %d ", id());
    //rd.print();  
    return true;
  }

  return false;
}

#endif // _View_change_h
