#include "View_info.h"
#include "View_change.h"
#include "View_change_ack.h"
#include "New_view.h"
#include "Pre_prepare.h"
#include "State.h"
#include "Status.h"
#include "Replica.h"
#include "Big_req_table.h"
#include "th_assert.h"
#include "K_max.h"
#include "Log.t"
#include "Array.t"

View_info::VCA_info::VCA_info() : v(0), vacks((View_change_ack *)0, node->n()) {}

void View_info::VCA_info::clear() {
  for (int i=0; i < node->n(); i++) {
    delete vacks[i];
    vacks[i] = 0;
  }
  v = 0;
}


View_info::View_info(int ident, View vi=0) :  v(vi), id(ident),
	   oplog(max_out), last_views((View)0, node->n()),
	   last_vcs((View_change*)0, node->n()),my_vacks((View_change_ack*)0, node->n()), 
	   vacks(node->n()), last_nvs(node->n())			      
	  	 
{
  last_stable = 0;
  last_committed = 0;
  last_executed = 0;
  vacks._enlarge_by(node->n());
  last_nvs._enlarge_by(node->n());
}


View_info::~View_info() {
  for (int i=0; i < last_vcs.size(); i++) {
    delete last_vcs[i];
    delete my_vacks[i];
  }
}

void View_info::add_chkpt_digest(Digest d) {
  cd = d;
}


void View_info::add_request(Order_request *pp, Seqno n) {
  th_assert(pp->view() == v, "Invalid argument");

  OReq_info &ri = oplog.fetch(pp->seqno());
  th_assert(pp->view() >= 0 && pp->view() > ri.lv, "Invalid argument");
  
  ri.clear();
  ri.lv = pp->view();
  ri.m = pp;
  ri.n = n;
  ri.d = pp->req_digest(n);
  ri.rd = pp->request_history_digest();
}

void View_info::discard_old() {
  // Discard view-changes, view-change acks, and new views with view
  // less than "v"
  for (int i=0; i < node->n(); i++) {
    if (last_vcs[i] && last_vcs[i]->view() < v) {
      delete last_vcs[i];
      last_vcs[i] = 0;
    }

    delete my_vacks[i];
    my_vacks[i] = 0;
    
    if (vacks[i].v < v) {
      vacks[i].clear();
      vacks[i].v = v;
    }

    if (last_nvs[i].view() < v) {
      last_nvs[i].clear();
    }
  }
}

void View_info::clear() {
  oplog.clear(last_stable+1);

  for (int i=0; i < node->n(); i++) {
    delete last_vcs[i];
    last_vcs[i] = 0;
    last_views[i] = v;
       
    last_nvs[i].clear();
  }
  vc_sent = zeroTime();
}

void View_info::view_change(View vi, Seqno ls, Seqno lc, Seqno le) {
  v = vi;
  discard_old();

  // Create my view-change message for "v".
  View_change *vc = new View_change(v, ls, lc, le, id);

  // Add checkpoint information
  vc->add_checkpoint(ls, cd, vi);
  fprintf(stderr,"VC : Adding checkpoint : \n");
  cd.print();
  
  /*fprintf(stderr, "Sending vc from %d with view: %llu digest: ", vc->id(),vc->view());
  Digest hd;
  Seqno lls;
  vc->ckpt(hd,lls);
  hd.print();*/

  // Add request information since last committed
  for (Seqno i=ls+1; i<=le; i++) {
    OReq_info &ri = oplog.fetch(i);
    
    if (ri.lv >= 0) {
      vc->add_request(i, ri.lv, ri.d, ri.rd);
    }
  }

  vc->authenticate();
  vc_sent = currentTime();
  fprintf(stderr, "Sending view change to other replicas \n");
  node->send(vc, Node::All_replicas);

  // Record that this message was sent.
  last_vcs[id] = vc;
  last_views[id] = v;
 
  int primv = node->primary(v);
  if (primv == id) {
    // If we are the primary create new view info for "v"
    NV_info &n = last_nvs[id];
    th_assert(n.view() <= v, "Invalid state");
    
    // Create an empty new-view message and add it to "n". Information
    // will later be added to "n/nv".
    New_view* nv = new New_view(v);
    n.add(nv, this);

    // Move any view-change messages for view "v" to "n".
    for (int i=0; i < node->n(); i++) {
      View_change* vc = last_vcs[i];
      if (vc && vc->view() == v && n.add(vc, true)) {
	last_vcs[i] = 0;
      }
    }

    // Move any view-change acks for messages in "n" to "n"
    for (int i=0; i < node->n(); i++) {
      VCA_info& vaci = vacks[i];
      if (vaci.v == v) {
	for (int j=0; j < node->n(); j++) {
	  if (vaci.vacks[j] && n.add(vaci.vacks[j])) 
	    vaci.vacks[j] = 0;
	}
      }
    }
  }
  else {
    // If we are not the primary, send view-change acks for messages in
    // last_vcs with view v.
    for (int i=0; i < node->n(); i++) {
      View_change *lvc = last_vcs[i];
      if (lvc && lvc->view() == v && i != id && i != primv) {
	View_change_ack *vack = new View_change_ack(v, id, i, lvc->digest());
	my_vacks[i] = vack;
	node->send(vack, primv);
      }
    }
  }
}

void View_info::mark_stable(Seqno ls) {
  last_stable = ls;
  oplog.truncate(last_stable+1);

  last_nvs[node->primary(v)].mark_stable(ls);
}

bool View_info::add(View_change* vc) {
  int vci = vc->id();
  int vcv = vc->view();

  if (vcv < v) {
    delete vc;
    return false;
  }
 
  // Try to match vc with a new-view message.
  NV_info &n = last_nvs[node->primary(vcv)];
  bool stored = false;
  int primv = node->primary(v);

  /*fprintf(stderr, "Recvd vc from %d with view: %llu digest: ", vc->id(),vc->view());
  Digest d;
  Seqno lls;
  vc->ckpt(d,lls);
  d.print();*/
	
  if (n.view() == vcv) {
    // There is a new-view message corresponding to "vc"
    bool verified = vc->verify();
    //fprintf(stderr, "Adding vc from %d is verified: %u \n", vc->id(),verified);
    stored = n.add(vc, verified);
    
    if (stored && id == primv && vcv == v) {
      // Try to add any buffered view-change acks that match vc to "n"
      for (int i=0; i < node->n(); i++) {
	VCA_info& vaci = vacks[i];
	if (vaci.v == v && vaci.vacks[vci] && n.add(vaci.vacks[vci])) {
	  vaci.vacks[vci] = 0;
	}
      }
    }    
  
    //fprintf(stderr, "Adding vc from %d stored :%u \n", vc->id(),stored);
    if (verified && vcv > last_views[vci]) { 
      last_views[vci] = vcv;
    }
  }
  else {
    // There is no matching new-view.
    if (vcv > last_views[vci] && vc->verify()) {
      delete last_vcs[vci];
      last_vcs[vci] = vc;
      last_views[vci] = vcv;
      stored = true;
    
      if (id != primv && vci != primv && vcv == v) { 
	// Send view-change ack.
	View_change_ack *vack = new View_change_ack(v, id, vci, vc->digest());
	th_assert(my_vacks[vci] == 0, "Invalid state");
	
	my_vacks[vci] = vack;
	node->send(vack, primv);
      }
    }  
  }

  if (!stored) 
    delete vc;

  return stored;
}

void View_info::add(View_change_ack* vca) {
  int vci = vca->vc_id();
  int vcv = vca->view();
  
  if (vca->verify()) {
    int primvcv = node->primary(vcv);

    NV_info &n = last_nvs[primvcv];
    if (n.view() == vcv && n.add(vca)) {
      // There is a new-view message corresponding to "vca"
      return;
    }

    if (id == primvcv) {
      VCA_info &vcai = vacks[vca->id()];
      if (vcai.v <= vcv) {
	if (vcai.v < vcv) {
	  vcai.clear();
	} 
	
	delete vcai.vacks[vci];
	vcai.vacks[vci] = vca;
	vcai.v = v;
	return;
      }
    }
  }

  delete vca;
}


void View_info::add(New_view* nv) {
  int nvi = nv->id();
  int nvv = nv->view();

  if (nvv >= v && nv->verify()) {
    NV_info &n = last_nvs[nvi];
    if (nv->view() > n.view()) {
      bool stored = n.add(nv, this);
      if (stored) {
	// Move any view-change messages for view "nvv" to "n".
	for (int i=0; i < last_vcs.size(); i++) {
          View_change* vc = last_vcs[i];
	  if (vc && vc->view() == nvv && n.add(vc, true)) {
	    last_vcs[i] = 0;
	  }
	}
      }
      return;
    }
  }

  delete nv;
}

inline View View_info::k_max(int k) const {
  return K_max<View>(k, last_views.as_pointer(), node->n(), View_max);
}

View View_info::max_view() const {
  View ret = k_max(node->f()+1);
  return ret;
}


View View_info::max_maj_view() const {
  View ret = k_max(node->n_f());
  return ret;
}

New_view *View_info::my_new_view(Time **t=0) {
  return last_nvs[id].new_view(t);
}


