#ifndef _View_info_h
#define _View_info_h 1

#include "Array.h"
#include "types.h"
#include "Digest.h"
#include "Node.h"
#include "NV_info.h"
#include "Log.h"
#include "Pre_prepare.h"

class View_change;
class New_view;
class State;
class View_change_ack;
class Status;
class Replica;
class Order_request;

class View_info {
  //
  // Holds information for the view-change protocol.
  //
public:
  View_info(int id, View v=0);
  // Effects: Create a view-info object for replica "id" with initial
  // view "v".
  
  ~View_info();
  // Effects: Discards all storage associated with this.
 

  // ZYZZYVA : TO DO 
   //
  // View-change creation:
   //

  void view_change(View v, Seqno ls, Seqno lc, Seqno le);
  // Creates view change message
  // Moves to view v, discards messages for views less than v.
  // ls - last stable, lc - last committed, lo -last ordered


  void add_chkpt_digest(Digest d);

  void add_request(Order_request *pp, Seqno n);
  // Add order request to the log
  
  bool fetch_history_digest(Seqno n, Digest &rhd);
  // Read request history digest of seqno n

  void add_missing(Order_request *pp);
  // TO DO: If Order request is missing at the local replica to 
  // complete a view change

  bool add(View_change* vc);
  // Effect: Stores message if it is valid and if it is not obsolete
  // otherwise discards it. This becomes the owner of "vc"'s storage.
  // Returns true iff "vc" is not discarded.

  void add(New_view* nv);
  // Effect: Stores message if it is valid and if it is not obsolete
  // otherwise discards it. This becomes the owner of "nv"'s storage.
   
  void add(View_change_ack* vca);
  // Effect: Adds message to this if valid and useful. 
 
  //
  // Observers:
  //
  View max_view() const;
  // Effects: Returns the maximum view number "v" for which it is
  // known that some correct replica is in view "v" or a later view.
  
  View max_maj_view() const;
  // Effects: Returns the maximum view number "v" for which it is
  // known that f+1 correct replicas are in view "v" or a later view.

   View view() const;
  // Effects: Returns the current view.

   
  bool has_new_view(View v) const;
  // Effects: Returns true iff this contains complete new-view
  // information for view v.

  bool has_nv_message(View v) const;
  // Effects: Returns true iff this contains a new-view message for
  // view v.

  View_change* my_view_change(Time **t=0);
  // Effects: Returns the view-change message sent by the calling replica
  // in view "view()".

  New_view* my_new_view(Time **t=0);
  // Effects: Returns any new-view message sent by the calling replica
  // in view "view()" or 0 if there is no such message.

   //
  // Misc.:
  //
  void mark_stable(Seqno ls);
  // Effects: Marks all requests with sequence number greater than or
  // equal to "ls" stable.

  void mark_stale();
  // Effects: Marks all the view-change and new-view messages in this
  // stale.  Except view-change messages sent by this or complete
  // new-view messages.

  void clear();
  // Effects: Removes all messages from this except that it retains log for
  // committed requests in old views.

 private:
  View v;
  int id;             // identifier of corresponding replica
  Seqno last_stable;  // last sequence number known to be stable
  Seqno last_committed; // last sequence number known to be committed locally
  Seqno last_executed; // last sequence number know to be ordered
  Digest cd;           // last stable checkpoint digest 
  
  struct OReq_info {
    View lv;
    Digest d;
    Digest rd; // history digest
    Seqno n;
    Order_request *m;  
    
    OReq_info() {lv = -1;}
    ~OReq_info(){};
    void clear(){lv = -1;};
  };
  
  //
  // oplog is a log of requests that were ordered/committed in 
  // previous views.
  //
  Log<OReq_info> oplog;

  // Highest view numbers in view-changes received from each replica.
  // It is indexed by replica id.
  Array<View> last_views;

  // View-change messages with the highest view number "vn" for each
  // replica such that "vn >= v" and there is no new-view message for
  // "vn". It is indexed by replica id.
  Array<View_change*> last_vcs; 
 
   // Keeps view-change acks sent in the current view by this replica.
  Array<View_change_ack*> my_vacks;

  // Buffered acknowledgments from other replicas.
  struct VCA_info {
    View v;
    Array<View_change_ack*> vacks;
    VCA_info();
    void clear();
  };
  Array<VCA_info> vacks;

  // New-view messages with the highest view number "vn" for each
  // replica (such that "vn >= v") and associated view-change
  // messages.  It is indexed by replica id.
  Array<NV_info> last_nvs;

  Time vc_sent; // time at which my view-change was last sent
  
  // 
  // Auxiliary methods:
  //

  void discard_old();
  // Effects: Discards messages with views lower than "v"

  View k_max(int k) const;
  // Effects: Returns the value "v" in "last_views" such that there
  // are "k" values greater than or equal to "v" in "last_views".
};

inline bool View_info::fetch_history_digest(Seqno n, Digest &d) {
  return last_nvs[node->primary(v)].fetch_history_digest(n, d);
}
  
inline View View_info::view() const {
  return v;
}

inline bool View_info::has_new_view(View vi) const {
  return last_nvs[node->primary(vi)].complete();
}

inline bool View_info::has_nv_message(View vi) const {
  if (last_nvs[node->primary(vi)].view())
    return true;
  return false;
}

inline void View_info::add_missing(Order_request *pp){  
  // TO DO: If Order request is missing at the local replica to 
  // complete a view change
}


#endif // _View_info_h
